from __future__ import annotations

from gmaps_route.nodes.node import Node
from gmaps_route.route.exceptions import RouteException
from gmaps_route.route.node_item import ABCNodeItem
from gmaps_route.route.route_point import RoutePoint


class Subdestination(ABCNodeItem, RoutePoint):

    class Builder:

        def __init__(self):
            self._latitude: Optional[float] = None
            self._longitude: Optional[float] = None

        def latitude(self, latitude: float) -> Subdestination.Builder:
            self._latitude = latitude
            return self

        def longitude(self, longitude: float) -> Subdestination.Builder:
            self._longitude = longitude
            return self

        def build(self) -> Subdestination:
            if self._latitude is None or self._longitude is None:
                raise RouteException("Latitude and longitude are required.")
            return Subdestination(self._latitude, self._longitude)

    def __init__(self, latitude: float, longitude: float):
        super(Subdestination, self).__init__(latitude, longitude)

        lat_lon_node = Node(1, [self.longitude, self.latitude])
        self._node = Node(3, [lat_lon_node])

    @property
    def node(self) -> Node:
        return self._node



from __future__ import annotations

import re

from abc import ABC, abstractmethod
from enum import Enum


class ABCNode(ABC):

    class ABCData(ABC):

        __data_regex = re.compile("^(\d+)([a-z])(.+)$")

        class DataType(str, Enum):

            # TODO: find out what these stand for - currently I have no idea
            D = "d"
            S = "s"
            M = "m"
            E = "e"
            B = "b"

            def __str__(self):
                return self.value

        def __init__(self, id: int, type: DataType):
            self._id = id
            self._type = type

        @property
        def id(self) -> int:
            return self._id

        @property
        def type(self) -> DataType:
            return self._type

        @property
        @abstractmethod
        def value(self) -> str:
            pass

    @property
    @abstractmethod
    def children(self) -> list[ABCNode]:
        pass

    @property
    @abstractmethod
    def data(self) -> ABCData:
        pass

    def has_children(self) -> bool:
        return len(self.children) > 0

    def add_child(self, child: ABCNode) -> None:
        self.children.append(child)

    def flatten(self) -> list[ABCNode]:
        def _flatten(node: ABCNode, nodes: list[ABCNode] = []) -> None:
            nodes.append(node)
            for child in node.children:
                _flatten(child, nodes)

        nodes: list[ABCNode] = []
        _flatten(self, nodes)
        return nodes

    def __str__(self) -> str:
        return f"{str(self.data.id)}{self.data.type}{self.data.value}"



from gmaps_route.nodes.abc_node import ABCNode
from gmaps_route.nodes.leaf import Leaf


class RoutePoint:
    def __init__(self, latitude: float, longitude: float):
        self._longitude = Leaf(
            1, ABCNode.ABCData.DataType.D, RoutePoint._format_coordinate(longitude)
        )
        self._latitude = Leaf(
            2, ABCNode.ABCData.DataType.D, RoutePoint._format_coordinate(latitude)
        )

    @staticmethod
    def _format_coordinate(coord: float) -> str:
        return f"{coord:.7f}"

    @property
    def latitude(self) -> Leaf:
        return self._latitude

    @property
    def longitude(self) -> Leaf:
        return self._longitude



from __future__ import annotations

from gmaps_route.nodes.abc_node import ABCNode


class Leaf(ABCNode):

    class LeafData(ABCNode.ABCData):

        def __init__(self, id: int, type: ABCNode.ABCData.DataType, value: str):
            super(Leaf.LeafData, self).__init__(id, type)
            self._value = value

        @property
        def value(self) -> str:
            return self._value

    def __init__(self, id: int, type: ABCNode.ABCData.DataType, value: str):
        self._children: list[ABCNode] = []
        self._data = Leaf.LeafData(id, type, value)

    @property
    def children(self) -> list[ABCNode]:
        return self._children

    @property
    def data(self) -> ABCNode.ABCData:
        return self._data



from __future__ import annotations

from gmaps_route.nodes.abc_node import ABCNode


class RootNode(ABCNode):

    class RootData(ABCNode.ABCData):

        @property
        def value(self) -> str:
            return ""

    def __init__(self, children: list[ABCNode] = []):
        self._children = children
        self._data = RootNode.RootData(-1, ABCNode.ABCData.DataType.M)

    @property
    def children(self) -> list[ABCNode]:
        return self._children

    @property
    def data(self) -> RootNode.RootData:
        return self._data

    def __str__(self) -> str:
        return ""



from __future__ import annotations

from enum import Enum

from gmaps_route.nodes.abc_node import ABCNode
from gmaps_route.nodes.leaf import Leaf
from gmaps_route.route.leaf_item import ABCLeafItem


class Transportation(ABCLeafItem):

    class Type(int, Enum):
        CAR = 0
        BIKE = 1
        FOOT = 2
        TRANSIT = 3
        FLIGHT = 4

    def __init__(self, transportation_type: Transportation.Type):
        self._transportation = Leaf(
            3, ABCNode.ABCData.DataType.E, str(int(transportation_type))
        )

    @property
    def leaf(self) -> Leaf:
        return self._transportation



from __future__ import annotations

from typing import Optional

from gmaps_route.nodes.abc_node import ABCNode
from gmaps_route.nodes.node import Node
from gmaps_route.route.exceptions import RouteException
from gmaps_route.route.node_item import ABCNodeItem
from gmaps_route.route.route_point import RoutePoint
from gmaps_route.route.subdestination import Subdestination


class Destination(ABCNodeItem, RoutePoint):

    class Builder:

        def __init__(self):
            self._latitude: Optional[float] = None
            self._longitude: Optional[float] = None
            self._subdestinations: list[Subdestination] = []

        def latitude(self, latitude: float) -> Destination.Builder:
            self._latitude = latitude
            return self

        def longitude(self, longitude: float) -> Destination.Builder:
            self._longitude = longitude
            return self

        def add_subdestination(
            self, subdestination: Subdestination
        ) -> Destination.Builder:
            self._subdestinations.append(subdestination)
            return self

        def build(self) -> Destination:
            if self._latitude is None or self._longitude is None:
                raise RouteException("Latitude and longitude are required.")
            return Destination(self._latitude, self._longitude, self._subdestinations)

    def __init__(
        self,
        latitude: float,
        longitude: float,
        subdestinations: list[Subdestination] = [],
    ):
        super(Destination, self).__init__(latitude, longitude)

        lat_lon_node = Node(2, [self.longitude, self.latitude])

        destination_nodes: list[ABCNode] = [lat_lon_node]
        for subdestination in reversed(subdestinations):
            destination_nodes.append(subdestination.node)

        self._node = Node(1, destination_nodes)

    @property
    def node(self) -> Node:
        return self._node



from __future__ import annotations

from typing import Optional

from gmaps_route.nodes.abc_node import ABCNode
from gmaps_route.nodes.leaf import Leaf
from gmaps_route.nodes.node import Node
from gmaps_route.nodes.root_node import RootNode
from gmaps_route.route.destination import Destination
from gmaps_route.route.exceptions import RouteException
from gmaps_route.route.map import Map
from gmaps_route.route.transportation import Transportation


class Route:

    GMAPS_DATA_URL = "https://www.google.com/maps/dir/data="

    class Builder:

        def __init__(self):
            self._map: Optional[Map] = None
            self._transportation: Optional[Transportation] = None
            self._destinations: list[Destination] = []

        def map(self, map: Optional[Map]) -> Route.Builder:
            self._map = map
            return self

        def transportation(
            self, transportation: Optional[Transportation]
        ) -> Route.Builder:
            self._transportation = transportation
            return self

        def add_destination(self, destination: Destination) -> Route.Builder:
            self._destinations.append(destination)
            return self

        def build(self) -> Route:
            return Route(self._destinations, self._transportation, self._map)

    def __init__(
        self,
        destinations: list[Destination],
        transportation: Optional[Transportation] = None,
        map: Optional[Map] = None,
    ):
        if len(destinations) < 2:
            raise RouteException("At least two destinations required.")

        self._root = RootNode()
        inner_route_node = Node(4)
        route_node = Node(4, [inner_route_node])

        if map is not None:
            self._root.add_child(map.node)

        for destination in reversed(destinations):
            inner_route_node.add_child(destination.node)

        if transportation is not None:
            inner_route_node.add_child(transportation.leaf)

        self._root.add_child(route_node)

    @property
    def data(self) -> str:
        nodes = self._root.flatten()
        return "!".join([str(n) for n in nodes])

    @property
    def url(self) -> str:
        return f"{Route.GMAPS_DATA_URL}{self.data}"



from __future__ import annotations

from gmaps_route.nodes.abc_node import ABCNode


class Node(ABCNode):

    class NodeData(ABCNode.ABCData):
        NODE_TYPE = ABCNode.ABCData.DataType.M

        def __init__(self, id: int, node: Node):
            super(Node.NodeData, self).__init__(id, Node.NodeData.NODE_TYPE)
            self._parent = node

        @property
        def value(self) -> str:
            return str(len(self._parent.flatten()) - 1)

    def __init__(self, id: int, children: list[ABCNode] = []):
        self._children = children
        self._data = Node.NodeData(id, self)

    @property
    def children(self) -> list[ABCNode]:
        return self._children

    @property
    def data(self) -> Node.NodeData:
        return self._data



from __future__ import annotations

from enum import Enum

from gmaps_route.nodes.abc_node import ABCNode
from gmaps_route.nodes.leaf import Leaf
from gmaps_route.nodes.node import Node
from gmaps_route.route.node_item import ABCNodeItem


class Map(ABCNodeItem):

    class Type(int, Enum):
        MAP = 1
        SATELLITE = 3

    def __init__(self, map_type: Map.Type):
        if map_type == Map.Type.SATELLITE:
            data_type = ABCNode.ABCData.DataType.E
            data_id = 1
        else:
            data_type = ABCNode.ABCData.DataType.B
            data_id = 4

        map_leaf = Leaf(data_id, data_type, str(int(map_type)))
        self._map = Node(3, [map_leaf])

    @property
    def node(self) -> Node:
        return self._map
